<copyright> Meta TOM DocBook output.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1999 Pieter J. Schoenmakers.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA

    <id>$Id: DBOutput.t,v 1.3 2000/09/22 03:39:40 bruce Exp $</id>
    </copyright>

implementation class
DBOutput: Output, Version

void
  outputFilePreamble String title
{
  [[out_file print ("<section><title>", title, "</title>")] nl];
}

void
  outputFilePostamble
{
  [[out_file print ("Generated by tm ", VERSION)] nl];
}

/******************** unit menu generation ********************/

void
  generateUnitMenu mu.Unit u
{
  Mapped dirs = [self relativeDirectoriesForUnit u];
  Indexed used = [u used];
  int i, n = !used ? 0 : [used length];
  Enumerator e;
  TomFile f;
  boolean b;

  [self outputFilePreamble [[MutableByteString new] print ("Unit ", [u name])]];

  if (n != 0)
    {
      MutableArray used_sorted = [MutableObjectArray withEnumerable used];
      mu.Unit builtin = [mu.Unit builtin_unit];
      boolean first = YES;

      [used_sorted bubbleSortUsingKey selector (String name)];

      for (i = 0; i < n; i++)
	{
	  mu.Unit dep = used_sorted[i];

	  if (dep != builtin)
	    {
	      if (first)
		{
		  [out_file print ("<section id=\"unit:", [u name],
				   "\"><title>Used units</title>\n",
				   "<para>\n")];
		  first = NO;
		}

	      [[out_file print (dirs[dep], " ", [dep name])] nl];
	    }
	}

      if (!first)
	[[out_file print "</para>"] nl];
    }

  [[out_file print "Files\n"] nl];

  MutableArray a = [MutableObjectArray withEnumerable [u files]];
  [a quickSortUsingKey selector (String name)];
  e = [a enumerator];

  while ({(b, f) = [e next]; b;})
    {
      MutableArray sorted = [MutableObjectArray withEnumerable [f elements]];
      [sorted quickSortUsingKey selector (String descriptive_name)];
      Enumerator ee = [sorted enumerator];

      String fn = [f name];
      mu.Extension x;

      [[out_file print ("<term>\"", fn, ".html\"", fn, "</term>")] nl];
      [[out_file print "<listitem>"] nl];

      while ({(b, x) = [ee next]; b;})
	{
	  String s = (![x name] ? [Meta ([[x meta] meta]) name]
		      : [x descriptive_name]);
	  [[out_file print ("<para>", s, "</para>")] nl];
	}

      [[out_file print "blurb"] nl];
    }

  [[out_file print "foop"] nl];

  [self outputFilePostamble];
}

end;

implementation instance
DBOutput
{
  <doc> Mapping from unit to directory name relative to the current unit.
      </doc>
  Mapped dirs;

  <doc> The level of the currently open heading.  </doc>
  int open_section;

  <doc> The number of open headings.  </doc>
  int open_section_num;

  <doc> Flag indicating whether we're in a method list.  </doc>
  boolean in_methods;

  <doc> When we are in a method list, flag indicating the next doc is a
      method doc.  </doc>
  boolean next_doc_is_method_doc;

  <doc> Set while doing meta variables.  Any documentation not
      specifically atatched to a declaration forces a break in the
      variable list.  </doc>
  boolean in_meta_vars;

  <doc> During meta variables, if set, the closing TERM has not yet been
      emitted.  This can is emitted early if the meta variable carries
      documentation.  This is set during documentation to indicate that it
      is meta-variable documentation.  </doc>
  boolean meta_var_pending;

  <doc> If set, a section has been openened but not yet filled with
      anything.  </doc>
  boolean section_empty;

  <doc> If set, at least one super was listed.  </doc>
  boolean have_supers;
}

id
  init ParseTom p
{
  TomFile f = [p file];

  [isa outputFilePreamble [[MutableByteString new]
			    print ("File ", [[f unit] name], '/', [f name])]];

  dirs = [isa relativeDirectoriesForUnit [f unit]];

  = [super init p];
}

void
  closeSection
{
  if (section_empty)
    {
      [out_file print "<para></para>"];
      section_empty = FALSE;
    }
  [[out_file print "</section>"] nl];
}

void
  finishFile
{
  while (open_section_num >= 0)
    {
      [self closeSection];
      open_section_num--;
    }
}

/******************** Output methods ********************/

void
  docCode String value
{
  [out_file print ("<literal>", value, "</literal>")];
}

void
  docPar
{
  [[out_file nl] print "</para><para>"];
}

void
  metaVarDocumentation String s
{
  if (meta_var_pending)
    [out_file print "</literal></term>"];

  [out_file print "<listitem>"];
  [super metaVarDocumentation s];
  [out_file print "</listitem>"];

  meta_var_pending = FALSE;
}

void
  finishMethods
{
  if (in_methods)
    in_methods = FALSE;
}

void
  startMethodDecl
{
  if (!in_methods)
    {
      in_methods = TRUE;
      section_empty = FALSE;
    }

  [out_file print "<programlisting>\n"];
}

void
  finishMethodDecl
{
  [[out_file print "</programlisting>"] nl];
}

void
  documentation String s
{
  if (in_methods && !next_doc_is_method_doc)
    {
      [out_file print "<para>INMETHODSDOC</para>"];
      in_methods = FALSE;
    }
  else if (in_meta_vars && !meta_var_pending)
    {
      [out_file print "</variablelist>"];
      in_meta_vars = FALSE;
    }

  next_doc_is_method_doc = FALSE;
  [out_file print "<para>"];
  [super documentation s];
  [[out_file print "</para>"] nl];
}

void
  methodDocumentation String s
{
  next_doc_is_method_doc = TRUE;
  [super methodDocumentation s];
}

void
  startHeading int i
{
<0>
  if (open_section >= i)
    {
      [out_file print "</section>"];
      open_section_num--;
    }
</0>
  if (i < 3)
    {
      [out_file print "<section><title>"];
      section_empty = TRUE;
      open_section_num++;
      open_section = i;
    }
  else
    [out_file print "<para><emphasis>"];
}

void
  finishHeading int i
{
  if (i < 3)
    [[out_file print "</title>"] nl];
  else
    [out_file print "</emphasis></para>"];
}

void
  startMetaVars
{
  section_empty = FALSE;
  [super startMetaVars];
}

void
  outputMetaVarStart
{
  if (!in_meta_vars)
    {
      [[out_file print "<variablelist>"] nl];
      in_meta_vars = TRUE;
    }
  [out_file print "<varlistentry><term><literal>"];
  meta_var_pending = TRUE;
}

void
  outputMetaVarEnd
{
  if (meta_var_pending)
    [out_file print "</literal></term>"];
  [out_file print "</varlistentry>"];
  meta_var_pending = FALSE;
}

void
  metaVar (int, Type, String) (qualifiers, type, name)
{
  // Mon Nov 22 21:00:35 1999, tiggr@gerbil.org
  // [self var-or-constant |{ || [super metaVar (qualifiers, type, name)]; }| ];
  [self outputMetaVarStart];
  [super metaVar (qualifiers, type, name)];
  [self outputMetaVarEnd];
}

void
  constant String c
     value Expression e
{
  [self outputMetaVarStart];
  [super constant c value e];
  [self outputMetaVarEnd];
}

void
  finishMetaVars
{
  if (in_meta_vars)
    {
      in_meta_vars = FALSE;
      [[out_file print "</variablelist>"] nl];
    }
}

OutputStream
  outputExtensionStart mu.Extension ext
{
  Meta m = [ext meta];
  String s = [m name];
  TomFile f = [[m extensionNamed nil] file];
  mu.Unit u = [f unit];

  if (![ext name])
    if ([m isClass])
      = [out_file print ("<link linkend=\"instance:", s, "\">", s, "</link>")];
    else
      = [out_file print ("<link linkend=\"class:", s, "\">", s, "</link>")];
  else
    = [out_file print ("<link linkend=\"class:", [m name], "\">", [m name],
		       "</link> (", [ext name], ")")];
}

OutputStream
  outputType Type t
{
  if (![t isMeta])
    = [out_file print ("<type>", [t name], "</type>")];
  else
    {
      Meta m = Meta (t);
      mu.Extension x = [m extensionNamed nil];
      boolean class_p = [m isClass];
      TomFile f = [x file];

      if (class_p)
	[out_file print "class ("];

      if (f != nil)
	= [out_file print ("<link linkend=\"instance:", [m name], "\">",
                           "<classname>", [m name], "</classname></link>")];
      else
	= [out_file print ("<classname>", [m name], "</classname>")];

      if (class_p)
	= [out_file print ')'];
    }
}

OutputStream
  outputUnit mu.Unit u
{
  String s = [u name];
  = [out_file print ("<link linkend=\"unit:", s, "\">", s, "</link>")];
}

void
  startOutputSupers
{
  section_empty = FALSE;
}

void
  outputFirstStateSuper Meta m
{
  have_supers = TRUE;
  [out_file print "<para>State supers: "];
  [super outputFirstStateSuper m];
}

void
  outputFirstBehaviourSuper Meta m
{
  if (have_supers)
    [out_file print "</para>"];
  else
    have_supers = TRUE;
  [out_file print "<para>Behaviour supers: "];
  [super outputFirstBehaviourSuper m];
}

void
  finishOutputSupers
{
  if (have_supers)
    {
      [[out_file print "</para>"] nl];
      have_supers = FALSE;
    }
}

redefine void
  startExtensionHeading mu.Extension x
{
  String xname = [x name];
  Meta m = [x meta];
  String mname = [m name];
  String mkind = [m isClass] ? "class" : "instance";

  if (open_section == 2)
    {
      [self closeSection];
      open_section_num--;
    }

  [out_file print ("<bridgehead id=\"", mkind, ":", mname)];
  if (xname != nil)
    [out_file print (":", xname)];
  [out_file print "\" renderas=\"sect2\">"];
}

void
  finishExtensionHeading mu.Extension x
{
  [out_file print "</bridgehead>"];
}

end;
