<copyright> Meta TOM HTML output.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1996, 1997 Pieter J. Schoenmakers.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA

    <id>$Id: HTMLOutput.t,v 1.28 2000/11/05 21:42:35 bruce Exp $</id>
    </copyright>

<doc> {HTMLOutput} is an {Output} for HTML format documentation.  </doc>
implementation class
HTMLOutput: Output, Version

void
  outputFilePreamble String title
{
  [[out_file print "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\" \"http://www.w3.org/TR/REC-html40/loose.dtd\">"] nl];
  [[out_file print ("<html><head><title>", title, "</title></head>")] nl];
  [[out_file print ("<body BGCOLOR=\"#ffffff\" TEXT=\"#000000\"",
		    " LINK=\"#000000\" VLINK=\"#000000\",
                    " ALINK=\"#000000\">")] nl];
  [[out_file print ("<h1>", title, "</h1>")] nl];
}

void
  outputFilePostamble
{
  [[out_file print ("<hr><address>Generated by tm ", VERSION,
		    ".</address></body></html>")] nl];
}

/******************** unit menu generation ********************/

void
  generateUnitMenu mu.Unit u
{
  Mapped dirs = [self relativeDirectoriesForUnit u];
  Indexed used = [u used];
  int i, n = !used ? 0 : [used length];
  Enumerator e;
  TomFile f;
  boolean b;

  [self outputFilePreamble [[MutableByteString new] print ("Unit ", [u name])]];

  if (n != 0)
    {
      MutableArray used_sorted = [MutableObjectArray withEnumerable used];
      mu.Unit builtin = [mu.Unit builtin_unit];
      boolean first = YES;

      [used_sorted bubbleSortUsingKey selector (String name)];

      for (i = 0; i < n; i++)
	{
	  mu.Unit dep = used_sorted[i];

	  if (dep != builtin)
	    {
	      if (first)
		{
		  [[out_file print "<h2>Used units</h2>\n<ul>"] nl];
		  first = NO;
		}

	      [[out_file print ("<li><a href=\"", dirs[dep], "index.html\">",
				[dep name], "</a>")] nl];
	    }
	}

      if (!first)
	[[out_file print "</ul>"] nl];
    }

  [[out_file print "<h2>Files</h2>\n<dl>"] nl];

  MutableArray a = [MutableObjectArray withEnumerable [u files]];
  [a quickSortUsingKey selector (String name)];
  e = [a enumerator];

  while ({(b, f) = [e next]; b;})
    {
      MutableArray sorted = [MutableObjectArray withEnumerable [f elements]];
      [sorted quickSortUsingKey selector (String descriptive_name)];
      Enumerator ee = [sorted enumerator];

      String fn = [f name];
      mu.Extension x;

      [[out_file print ("<dt><a href=\"", fn, ".html\">", fn, "</a>")] nl];
      [[out_file print "<dd><ul>"] nl];

      while ({(b, x) = [ee next]; b;})
	{
	  String s = (![x name] ? [Meta ([[x meta] meta]) name]
		      : [x descriptive_name]);
	  [[out_file print ("<li>", s)] nl];
	}

      [[out_file print "</ul>"] nl];
    }

  [[out_file print "</dl>"] nl];

  [self outputFilePostamble];
}

end;

implementation instance
HTMLOutput
{
  <doc> Mapping from unit to directory name relative to the current unit.
      </doc>
  Mapped dirs;

  <doc> Flag indicating whether we're in a method list.  </doc>
  boolean in_methods;

  <doc> When we're in a method list, flag indicating the next doc is a
      method doc.  </doc>
  boolean next_doc_is_method_doc;
}

id
  init ParseTom p
{
  TomFile f = [p file];

  [isa outputFilePreamble [[MutableByteString new]
			    print ("File ", [[f unit] name], '/', [f name])]];

  dirs = [isa relativeDirectoriesForUnit [f unit]];

  = [super init p];
}

void
  finishFile
{
  [isa outputFilePostamble];
}

/******************** Output methods ********************/

void
  startDocIdentifier
{
  [out_file print "<b><code>"];
}

void
  finishDocIdentifier
{
  [out_file print "</code></b>"];
}

void
  docCode String value
{
  [out_file print ("<code>", value, "</code>")];
}

void
  docPar
{
  [[out_file nl] print "<p>"];
}

void
  startExtensionHeading mu.Extension x
{
  if ([[x meta] isClass])
    [out_file print "<hr>"];
  [super startExtensionHeading x];
}

void
  metaVarDocumentation String s
{
  [out_file print "</code></b>\n<dd>"];
  [super metaVarDocumentation s];
}

void
  finishMethods
{
  if (in_methods)
    {
      in_methods = FALSE;
      [out_file print "</dl>"];
    }
}

void
  startMethodDecl
{
  if (!in_methods)
    {
      [out_file print "<dl>"];
      in_methods = TRUE;
    }
  [out_file print "<dt><b><pre>"];
}

void
  finishMethodDecl
{
  [[out_file print "</pre></b><dd>"] nl];
}

void
  documentation String s
{
  if (in_methods && !next_doc_is_method_doc)
    {
      [out_file print "</dl>"];
      in_methods = FALSE;
    }
  next_doc_is_method_doc = FALSE;
  [super documentation s];
}

void
  methodDocumentation String s
{
  next_doc_is_method_doc = TRUE;
  [super methodDocumentation s];
  [out_file print "<p>"];
}

void
  startHeading int i
{
  [out_file print ("<h", i, '>')];
}

void
  finishHeading int i
{
  [[out_file print ("</h", i, '>')] nl];
}

void
  startMetaVars
{
  [super startMetaVars];
  [[out_file print "<dl>"] nl];
}

void
  metaVar (int, Type, String) (qualifiers, type, name)
{
  [[out_file print "<dt><b><code>"] nl];
  [super metaVar (qualifiers, type, name)];
}

void
  constant String c
     value Expression e
{
  [[out_file print "<dt><b><code>"] nl];
  [super constant c value e];
}

void
  finishMetaVars
{
  [[out_file print "</dl>"] nl];
  [super finishMetaVars];
}

OutputStream
  outputExtensionStart mu.Extension ext
{
  Meta m = [ext meta];
  String s = [m name];
  TomFile f = [[m extensionNamed nil] file];
  mu.Unit u = [f unit];

  if (![ext name])
    if ([m isClass])
      = [out_file print ("<a name=\"", s, "\" href=\"#i_", s, "\">",
			 s, "</a>")];
    else
      = [out_file print ("<a name=\"i_", s, "\" href=\"#", s, "\">",
			 s, "</a>")];
  else
    = [out_file print ("<a href=\"", dirs[u], [f name], ".html#", [m name],
		       "\">", [m name], "</a> (", [ext name], ")")];
}

OutputStream
  outputType Type t
{
  if (![t isMeta])
    = [out_file print [t name]];
  else
    {
      Meta m = Meta (t);
      mu.Extension x = [m extensionNamed nil];
      boolean class_p = [m isClass];
      TomFile f = [x file];

      if (class_p)
	[out_file print "class ("];

      if (f != nil)
	if ([f unit] == [mu.Unit named [tm unit]])
	  = [out_file print ("<a href=\"", [f name], ".html#i_", [m name], "\">",
			     [m name], "</a>")];
	else
	  {
	    mu.Unit u = [f unit];

	    = [out_file print ("<a href=\"", dirs[u], "index.html\">",
			       [u name], "</a>.<a href=\"", dirs[u], [f name],
			       ".html#i_", [m name], "\">", [m name], "</a>")];
	  }
      else
	= [out_file print [m name]];

      if (class_p)
	= [out_file print ')'];
    }
}

OutputStream
  outputUnit mu.Unit u
{
  = [out_file print ("<a href=\"", dirs[u], "index.html\">", [u name], "</a>")];
}

void
  startOutputSupers
{
  [[out_file print "<dl>"] nl];
}

void
  outputFirstStateSuper Meta m
{
  [out_file print "<dt>State supers\n<dd>"];
  [super outputFirstStateSuper m];
}

void
  outputFirstBehaviourSuper Meta m
{
  [out_file print "<dt>Behaviour supers\n<dd>"];
  [super outputFirstBehaviourSuper m];
}

void
  finishOutputSupers
{
  [[out_file print "</dl>"] nl];
}

end;
