<copyright> Meta TOM abstract output.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1996, 1997 Pieter J. Schoenmakers.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA

    <id>$Id: Output.t,v 1.36 2000/11/19 08:29:14 bruce Exp $</id>
    </copyright>

implementation class
Output: State, tm
{
  <doc> Mapping from unit to mapping from units to relative directories.
      </doc>
  static MutableDictionary unit_dirs;

  <doc> The current output instance.  </doc>
  public static Output current;
}

<doc> Return the {Mapped} {Collection} from unit to relative directory, as
    seen from the directory of unit {u}.  </doc>
Mapped
  relativeDirectoriesForUnit mu.Unit u
{
  MutableDictionary d;

  if (!unit_dirs)
    unit_dirs = [MutableDictionary new];
  else
    d = unit_dirs[u];

  if (!d)
    {
      Enumerator eu = [[mu.Unit units] enumerator];
      String dir = [u directory];
      mu.Unit other;

      d = [MutableDictionary new];

      while ({boolean b; (b, other) = [eu next]; b;})
	d[other] = [File express-filename [other directory] relative-to dir];

      unit_dirs[u] = d;
    }

  = d;
}

<doc> Generate the unit menu.  This should be implemented by subclasses.
    </doc>
void
  generateUnitMenu mu.Unit u
{
  [self unimplemented cmd];
}

end;

implementation instance
Output
{
  ParseTom parser;

  <doc> The current extension being scrutinized.  </doc>
  mu.Extension ext;

  <doc> The documentation parser.  </doc>
  ParseDoc doc_parser;

  <doc> The documentation lexer.  </doc>
  LexDoc doc_lexer;
}

id
  init ParseTom p
{
  current = self;
  parser = p;

  = self;
}

void
  doSupers Indexed supers
{
  Meta m = [ext meta];
  Indexed ss = [m state_supers];
  Indexed bs = [m behaviour_supers];
  boolean first, is_class = [m isClass];
  int i, n;

  [self startHeading 3];
  [out_file print "inherits"];
  [self finishHeading 3];

  [self startOutputSupers];

  /* Output the state supers.  */
  if (ss != nil)
    {
      for ({first = YES; i = 0; n = [ss length];}; i < n; i++)
	{
	  Meta sup = ss[i];
	
	  if (is_class)
	    sup = [Class (sup) instance]; 

	  if (first)
	    {
	      [self outputFirstStateSuper sup];
	      first = NO;
	    }
	  else
	    [self outputStateSuper sup];
	}
      if (!first)
	[self finishOutputStateSupers];
    }

  if (bs != nil)
    {
      /* Output the behaviour supers.  */
      for ({first = YES; i = 0; n = [bs length];}; i < n; i++)
	if (first)
	  {
	    [self outputFirstBehaviourSuper bs[i]];
	    first = NO;
	  }
	else
	  [self outputBehaviourSuper bs[i]];
      if (!first)
	[self finishOutputBehaviourSupers];
    }

  [self finishOutputSupers];
}

void
  finishFile
{
  [self documentation [self savedDocumentation]];
}

void
  finishMethod CustomMethod m
{
  Indexed np = [m name_parts], args = [m arguments];
  int i, j, l, n, nnp = [np length], nap = !args ? 0 : [args length];

  [self startMethodDecl];
  [self outputQualifiers [m qualifiers]];
  [[m return_type] outputType self to out_file];

  if ([m return_names] != nil)
    [[m return_names] write [out_file print ' ']];

  for (i = 0; i < nnp; i++)
    {
      n = [np[i] length];
      if (n > l)
	l = n;
    }

  for ({l += 2; i = 0;}; i < nnp; i++)
    {
      String s = np[i];
      n = [np[i] length];

      [out_file print '\n'];
      for (j = n; j < l; j++)
	[out_file print ' '];
      [self outputMethodNamePart s];
      if (i < nap)
	{
	  Argument arg = args[i];

	  [out_file print ' '];
	  [arg outputType self to out_file];
	  [out_file print ' '];
	  [arg outputName self to out_file];

	  Expression e = [arg default_value];
	  if (e)
	    [out_file print (" = ", e)];
	}
    }

  Expression e = [m precondition];
  if (e != nil)
    [out_file print ("\npre\n  ", e)];

  e = [m postcondition];
  if (e != nil)
    [out_file print ("\npost\n  ", e)];
  
  [out_file print ";\n"];

  [self finishMethodDecl];

  [self methodDocumentation [self savedDocumentation]];
}

void
  outputMethodNamePart String s
{
  [out_file print s];
}

void
  metaKind
{
  [out_file print [[ext meta] isClass] ? "class" : "instance"];
}

String
  savedDocumentation
{
  LexTom lx = LexTom ([parser lex]);
  String s = [lx documentation];

  if (s != nil)
    [lx set_documentation nil];

  = s;
}

void
  startExtension mu.Extension x
       interface boolean is_interface
{
  ext = x;

  [self startExtensionHeading x];
  [self metaKind];
  [out_file print ' '];

  x = [[ext meta] extensionNamed nil];
  [self outputUnit [[x file] unit]];
  [out_file print '.'];

  [self outputExtensionStart ext];

  [self finishExtensionHeading x];

  if (![parser doc_method_p])
    [self documentation [self savedDocumentation]];
}

void
  startMetaVars
{
  [self startHeading 3];
  [out_file print "variables"];
  [self finishHeading 3];
}

void
  finishMetaVars
{
}

void
  startMethods
{
  [self startHeading 3];
  [out_file print "methods"];
  [self finishHeading 3];
}

void
  finishMethods
{
}

void
  constant String c
     value Expression e
{
  [out_file print "const "];
  [out_file print c];
  [[out_file print (" = ", e, ';')] nl];

  [self metaVarDocumentation [self savedDocumentation]];
}

void
  metaVar (int, Type, String) (qualifiers, type, name)
{
  [self outputQualifiers qualifiers];
  [self outputType type];
  [out_file print ' '];
  [self outputName name];
  [out_file print ';'];

  [self metaVarDocumentation [self savedDocumentation]];
}

protected boolean
  outputSingleQualifier String q
		  first boolean first
{
  if (!first)
    [out_file print ' '];
  [out_file print q];
  = NO;
}

void
  outputQualifiers int q
{
  boolean first = YES;

  if ([Qualified isDeferred q])
    first = [self outputSingleQualifier "deferred" first first];
  if ([Qualified isLocal q])
    first = [self outputSingleQualifier "local" first first];
  if ([Qualified isStatic q])
    first = [self outputSingleQualifier "static" first first];
  //  if ([Qualified isExtern q])
  //    first = [self outputSingleQualifier "extern" first first];
  if ([Qualified isPublic q])
    first = [self outputSingleQualifier "public" first first];
  if ([Qualified isPrivate q])
    first = [self outputSingleQualifier "private" first first];
  if ([Qualified isProtected q])
    first = [self outputSingleQualifier "protected" first first];
  if ([Qualified isMutable q])
    first = [self outputSingleQualifier "mutable" first first];
  if ([Qualified isRedeclare q])
    first = [self outputSingleQualifier "redeclare" first first];
  if ([Qualified isRedefine q])
    first = [self outputSingleQualifier "redefine" first first];
  if ([Qualified isObsolete q])
    first = [self outputSingleQualifier "obsolete" first first];

  if (!first)
    [out_file print ' '];
}

/******************** ParseDoc output methods ********************/

void
  startDoc
{
}

void
  finishDoc
{
  [out_file nl];
}

void
  docText String value
{
  [out_file print value];
}

void
  docCode String value
{
  [out_file print value];
}

void
  startDocIdentifier
{
}

void
  finishDocIdentifier
{
}

void
  docIdentifier String value
{
  Meta m = [[mu.Unit named [tm unit]] metaNamed value recurse: YES];

  if (!m)
    [self docCode value];
  else
    {
      [self startDocIdentifier];
      [m outputType self to out_file];
      [self finishDocIdentifier];
    }
}

void
  docPar
{
  [[out_file nl] nl];
}

/******************** Output methods ********************/

void
  documentation String s
{
  if (s != nil)
    {
      if (!doc_parser)
	{
	  doc_lexer = [LexDoc alloc];
	  doc_parser = [[ParseDoc alloc] init doc_lexer output self];
	}

      [doc_lexer initWithFilename [[parser lex] filename]
		 line [[parser lex] current_line] doc s];

      [doc_parser documentation];
    }
}

void
  methodDocumentation String s
{
  if (!s)
    s = "Undocumented.";
  [self documentation s];
}

void
  metaVarDocumentation String s
{
  [self documentation s];
  [out_file nl];
}

void
  startHeading int i
{
}

void
  finishHeading int i
{
  [out_file nl];
}

void
  startMethodDecl
{
}

void
  finishMethodDecl
{
}

OutputStream
  outputName String s
{
  = [out_file print s];
}

OutputStream
  outputExtensionStart mu.Extension ext
{
  = [out_file print [[ext meta] name]];
  if ([ext name] != nil)
    [out_file print (" (", [ext name], ')')];
}

OutputStream
  outputType Type t
{
  if (![t isMeta])
    = [out_file print [t name]];
  else
    {
      Meta m = Meta (t);
      mu.Extension x = [m extensionNamed nil];
      boolean class_p = [m isClass];
      TomFile f = [x file];

      if (class_p)
	[out_file print "class ("];

      Unit u = !f ? nil : [f unit];
      if (f != nil && (u != [Unit builtin_unit]
		       && u != [mu.Unit named [tm unit]]))
	= [out_file print ([u name], '.', [m name])];
      else
	= [out_file print [m name]];

      if (class_p)
	= [out_file print ')'];
    }
}

OutputStream
  outputUnit mu.Unit u
{
  = [out_file print [u name]];
}

void
  startOutputSupers
{
}

void
  outputFirstStateSuper Meta m
{
  [self outputType m];
}

void
  outputStateSuper Meta m
{
  [out_file print ", "];
  [self outputType m];
}

void
  finishOutputStateSupers
{
  [out_file nl];
}

void
  outputFirstBehaviourSuper Meta m
{
  [self outputType m];
}

void
  outputBehaviourSuper Meta m
{
  [out_file print ", "];
  [self outputType m];
}

void
  finishOutputBehaviourSupers
{
  [out_file nl];
}

void
  finishOutputSupers
{
}

void
  startExtensionHeading mu.Extension x
{
  [self startHeading 2];
}

void
  finishExtensionHeading mu.Extension x
{
  [self finishHeading 2];
}

end;
