<copyright> Meta TOM Texinfo output.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1996, 1997 Pieter J. Schoenmakers.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA

    <id>$Id: TexiOutput.t,v 1.13 1998/07/21 15:01:00 tiggr Exp $</id>
    </copyright>

<doc> {TexiOutput} is an {Output} for Texinfo format documentation.  </doc>
implementation class
TexiOutput: Output

<doc> Return the node name for the extension {x}.  </doc>
String (node_name)
  nodeNameForExtension mu.Extension x
{
  node_name = ![x name] ? [[x meta] name] : [x descriptive_name];
}

/******************** unit menu generation ********************/

<doc> Write the so-called unit menu to the {out_file}, which is
    {index.texi} in our case.  This only emits the proper {@include}
    directives.  </doc>
redefine void
  generateUnitMenu mu.Unit u
{
  Mapped dirs = [self relativeDirectoriesForUnit u];
  Indexed used = [u used];
  int i, n = !used ? 0 : [used length];
  Enumerator e;
  TomFile f;
  boolean b;

  MutableArray files = [MutableObjectArray withEnumerable [u files]];
  MutableArray file_contents = [MutableObjectArray withCapacity [files length]];
  MutableArray all_sorted = [MutableObjectArray new];

  [files quickSortUsingKey selector (String name)];
  e = [files enumerator];

  while ({(b, f) = [e next]; b;})
    {
      MutableArray sorted = [MutableObjectArray withEnumerable [f elements]];
      [sorted quickSortUsingKey selector (String descriptive_name)];
      [all_sorted addElementsFrom sorted];
      [file_contents add sorted];
    }

<0>
  [[out_file print ("\\input texinfo\n@setfilename ", [u name],
		    ".info\n@settitle Unit ", [u name])] nl];
  if ([File file-exists "preamble.texi"])
    [out_file print "@include preamble.texi\n"];
  [out_file print "@node Top\n@ifinfo\n@top Class Menu\n"];
</0>

  File menu_file = [File open "tm-menu.texi" output: YES
			 flags: FILE_NOT_EXIST_CREATE | FILE_EXIST_TRUNCATE];
  [menu_file print "@menu\n"];
  [all_sorted quickSortUsingKey selector (String descriptive_name)];
  mu.Extension x;
  e = [all_sorted enumerator];
  while ({(b, x) = [e next]; b;})
    {
      x = [[[x meta] itsClass] extensionNamed [x name]];
      [menu_file print ("* ", [self nodeNameForExtension x], "::\n")];
    }
  [menu_file print "@end menu\n"];
  [menu_file close];

  e = [files enumerator];
  while ({(b, f) = [e next]; b;})
    [[out_file print ("@include ", [f name], ".texi")] nl];

<0>
  if ([File file-exists "postamble.texi"])
    [out_file print "@include postamble.texi\n"];

  [[out_file print "@bye"] nl];
</0>
}

end;

implementation instance
TexiOutput

/******************** Output methods ********************/

redefine void
  startDocIdentifier
{
  [out_file print "@code{"];
}

redefine void
  finishDocIdentifier
{
  [out_file print "}"];
}

redefine void
  docCode String value
{
  [out_file print ("@code{", value, "}")];
}

redefine void
  metaVarDocumentation String s
{
  [[out_file print "\n@noindent"] nl];
  [super metaVarDocumentation s];
}

redefine void
  startMethodDecl
{
  [[out_file print "\n@sp 1\n@example\n@group"] nl];
}

redefine void
  finishMethodDecl
{
  [[out_file print "@end group\n@end example\n@noindent"] nl];
}

redefine void
  startHeading int i
{
  [out_file print i == 2 ? "@section " : "\n@subheading "];
}

redefine void
  startExtensionHeading mu.Extension x
{
  if ([[x meta] isClass])
    [out_file print ("\n@page\n@node ", [isa nodeNameForExtension x],
		     ", , , Unit ", [[[x file] unit] name], "\n")];
  [out_file print "@unnumberedsec "];
}

redefine void
  finishHeading int i
{
  [out_file print "\n@noindent\n"];
}

redefine void
  startMetaVars
{
  [super startMetaVars];
  [[out_file print "@table @code"] nl];
}

redefine void
  metaVar (int, Type, String) (qualifiers, type, name)
{
  [out_file print "@item "];
  [super metaVar (qualifiers, type, name)];
}

void
  constant String c
     value Expression e
{
  [out_file print "@item "];
  [super constant c value e];
}

redefine void
  finishMetaVars
{
  [[out_file print "@end table"] nl];
  [super finishMetaVars];
}

redefine void
  outputFirstStateSuper Meta m
{
  [out_file print "\n@noindent\nState supers: "];
  [self outputType m];
}

redefine void
  outputFirstBehaviourSuper Meta m
{
  [out_file print "\n@noindent\nBehaviour supers: "];
  [super outputFirstBehaviourSuper m];
}

end;
