<copyright> Meta TOM main.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1996-1998 Pieter J. Schoenmakers.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA

    <id>$Id: tm.t,v 1.46 2000/06/27 17:15:19 tiggr Exp $</id>
    </copyright>

implementation class
tm: stdio, Constants
{
  <doc> The current unit.  </doc>
  static public String unit;

  <doc> The current output file.  </doc>
  static OutputStream out_file;
}

<doc> Output usage information and exit.  </doc>
void
  usage
{
  [[err print ("usage: ", [Runtime program_name],
	       " -u unit [options] [-T|-f format [file]]
options:
 -o filename  send output to the file <filename>.<format> instead of stdout
 -f format    have output generated by the <format> class
 -fa          only report all units depended upon by the current unit
 -fA          like `-fa' but include the current unit
 -T           create an etags file (you wish!)
 -all         update all doc files (into the current dir)
 -I dir       include <dir> in the search path
 -public      don't print private methods
 -dd          trace the doc parser
 -dt          trace the TOM parser
 -du          trace the unit parser
where
 -all, -T, -fa, and -fA are incompatible with each other and the [file];
 -all is incompatible with -o;
 the default <format> class is HTML, other possibilities are Texi;
 the <format> used as extension for -o is lower cased.")] nl];
  [Runtime exit 2];
}

<doc> Set the {out_file} from the {basename} and {ext}ension.  </doc>
void
  set_out_file (String, String) (basename, ext)
pre
  !out_file
{
  if (!basename)
    out_file = out;
  else
    {
      String name = [[MutableByteString new] print (basename, ".", ext)];
      out_file = [File open name output: YES
		       flags: FILE_NOT_EXIST_CREATE | FILE_EXIST_TRUNCATE];
    }
}

<doc> Close the output file.  </doc>
void
  close_out_file
post
  !out_file
{
  [out_file close];
  out_file = nil;
}

<doc> The {main} function.  </doc>
int
  main Array arguments
{
  String out_filename, filename, format = "HTML";
  boolean make_tags, do_all, hide_privates;
  int i, n = [arguments length];
  int report_units;

  err = [BufferedStream with err];
  [Parser addToPath ""];

  for (i = 0; i < n; i++)
    {
      String s = arguments[i];

      if ([s equal "-u"])
	unit = arguments[++i];
      else if ([s equal "-f"])
	format = arguments[++i];
      else if ([s equal "-I"])
	[Parser addToPath arguments[++i]];
      else if ([s equal "-all"])
	do_all = YES;
      else if ([s equal "-du"])
	[ParseUnit set_trace_default YES];
      else if ([s equal "-dt"])
	[ParseTom set_trace_default YES];
      else if ([s equal "-dd"])
	[ParseDoc set_trace_default YES];
      else if ([s equal "-fa"])
	report_units = 1;
      else if ([s equal "-fA"])
	report_units = 2;
      else if ([s equal "-o"])
	out_filename = arguments[++i];
      else if ([s equal "-T"])
	{
	  make_tags = TRUE;
	  format = "TAGS";
	}
      else if ([s equal "-public"])
        hide_privates = TRUE;
      else if (s[0] == '-')
	[self usage];
      else if (!filename)
	filename = s;
    }

  if (!unit
      || (filename != nil && (make_tags || do_all || report_units > 0))
      || (make_tags && (do_all || report_units > 0))
      || (do_all && report_units > 0))
    [self usage];

  [Parser addToPath [[MutableByteString new]
		      print ([Runtime tom_prefix], "lib/tom/")]];
  [mu init FALSE];

  mu.Unit u = [mu.Unit named unit load: YES];
  if (!u)
    {
      [[[Parser error] print ("could not load unit named `", unit, '\'')] nl];
      return 1;
    }

  if (report_units != 0)
    {
      Enumerator e = [[mu.Unit units] enumerator];
      mu.Unit l, bi = [mu.Unit builtin_unit];
      boolean b, not_first;

      while ({(b, l) = [e next]; b;})
	{
	  if (l != bi && (l != u || report_units > 1))
	    {
	      if (not_first)
		[out print ' '];
	      else
		not_first = TRUE;
	      [out print [l name]];
	    }
	}

      [out nl];
      return;
    }

  String output_name = [[MutableByteString new] print (format, "Output")];
  class (Output) output_class
    = class (Output) ([Runtime classNamed output_name]);

  if (!output_class)
    {
      [[[Parser error] print ("No class named `", format, '\'')] nl];
      return 1;
    }

  format = [format downcase];

  if (!filename)
    if (!make_tags)
      {
	[self set_out_file (do_all ? !out_filename ? "index"
			    : out_filename : nil, format)];
	[output_class generateUnitMenu u];
	[self close_out_file];

	if (do_all)
	  {
	    Enumerator e = [[u files] enumerator];
	    TomFile f;
	    boolean b;

	    while ({(b, f) = [e next]; b;})
	      {
		[self set_out_file ([f name], format)];

		LexTom l = [[LexTom alloc]
			    initWithStream [File open [f filename] input: YES
						 flags: FILE_NOT_EXIST_RAISE]
			    named [f filename]];
		ParseTom pt = [ParseTom parser f using l];

		[pt set_output [[output_class alloc] init pt]];
                [pt set_hide_private_methods hide_privates];
		[pt parse];

		[self close_out_file];
		[l close];
	      }
	  }
      }
    else
      {
	Enumerator e = [[u files] enumerator];
	TAGSOutput out = [TAGSOutput new];
	TomFile f;
	boolean b;

	[self set_out_file ("TAGS", "")];

	while ({(b, f) = [e next]; b;})
	  {
	    LexTom l = [[LexTom alloc]
			initWithStream [File open [f filename] input: YES
					     flags: FILE_NOT_EXIST_RAISE]
			named [f filename]];

	    ParseTom pt = [ParseTom parser f using l];
	    [out set_parser pt];
	    [pt parse];
	  }

	[self close_out_file];
      }
  else
    {
      TomFile f = [u fileNamed filename];

      if (!f)
	{
	  [[[Parser error] print ("unit `", unit, "' has no file named `",
				  filename, '\'')] nl];
	  return 1;
	}

      [self set_out_file (!out_filename ? nil : out_filename, format)];

      LexTom l = [[LexTom alloc]
		  initWithStream [File open [f filename] input: YES
				       flags: FILE_NOT_EXIST_RAISE]
		  named [f filename]];

      ParseTom pt = [ParseTom parser f using l];

      [pt set_output [[output_class alloc] init pt]];
      [pt parse];

      [self close_out_file];
    }
}

end;

implementation instance
tm

//
deferred int compare All a;

end;

/******************** All ********************/

implementation class
All extension tm

end;

implementation instance
All extension tm

OutputStream
     write OutputStream s
  priority int p
{
  = [self write s];
}

end;

/******************** ArgumentTuple ********************/

<doc> Extension to {ArgumentTuple} for interfacing with {Output}.  </doc>
implementation class
ArgumentTuple extension tm
end;

implementation instance
ArgumentTuple extension tm

OutputStream
  outputType Output o
	  to OutputStream s
{
  int i;

  [s print '('];
  for (i = 0; i < length; i++)
    {
      if (i > 0)
	[s print ", "];
      [self[i] outputType o to s];
    }

  = [s print ')'];
}

OutputStream
  outputName Output o
	  to OutputStream s
{
  int i;

  [s print '('];
  for (i = 0; i < length; i++)
    {
      if (i > 0)
	[s print ", "];
      [self[i] outputName o to s];
    }

  = [s print ')'];
}

end;

/******************** Cast ********************/

implementation class
Cast extension tm
end;

implementation instance
Cast extension tm

redefine OutputStream
  write OutputStream s
{
  [[Output current] outputType desired_type];
  [s print (" (", expr, ")")];
}

end;

/******************** ClassReference ********************/

implementation class
ClassReference extension tm

end;

implementation instance
ClassReference extension tm

OutputStream
  write OutputStream s
{
  [[Output current] outputType [Meta (type) instance]];
}

end;

/******************** Common ********************/

implementation class
Common extension tm
end;

implementation instance
Common extension tm

id
  again
{
  = self;
}

OutputStream
  outputType Output o
	  to OutputStream s
{
  [self unimplemented cmd];
}

OutputStream
  outputName Output o
	  to OutputStream s
{
  [self unimplemented cmd];
}

end;

/******************** Expression ********************/

implementation class
Expression extension tm
end;

implementation instance
Expression extension tm

OutputStream
  outputType Output o
	  to OutputStream s
{
  = [o outputType [self type]];
}

end;

/******************** Named ********************/

implementation class
Named extension tm
end;

implementation instance
Named extension tm

OutputStream
  outputName Output o
	  to OutputStream s
{
  = [o outputName name];
}

OutputStream
  write OutputStream s
{
  = [s print name];
}

end;

/******************** NumberConstant ********************/

implementation class
NumberConstant extension tm

end;

implementation instance
NumberConstant extension tm

OutputStream
  write OutputStream s
{
  = [s print value];
}

end;

/******************** SelectorConstant ********************/

implementation class
SelectorConstant: Expression

instance (id)
  withNameParts MutableArray name_parts
	  types (Type, MutableArray) (return_type, arg_types)
{
  = [[self alloc] initWithNameParts name_parts types (return_type, arg_types)];
}

end;

implementation instance
SelectorConstant
{
  Indexed name_parts;
  Type return_type;
  Indexed arg_types;
}

protected id
  initWithNameParts Indexed np
	      types (Type, Indexed) (rt, at)
{
  (name_parts, return_type, arg_types) = (np, rt, at);
  = [super initWithType nil];
}

redefine OutputStream (s)
  write OutputStream s
{
  [s print "selector ("];
  [return_type write s];
  int i, nn = [name_parts length], na = !arg_types ? 0 : [arg_types length];
  for (i = 0; i < nn; i++)
    {
      [name_parts[i] write [s print ' ']];
      if (i < na)
	[arg_types[i] write [s print ' ']];
    }
  [s print ")"];
}

end;

/******************** Tuple ********************/

implementation class
Tuple extension tm

end;

implementation instance
Tuple extension tm

redefine OutputStream
  write OutputStream s
{
  int i;

  [s print '('];
  for (i = 0; i < length; i++)
    {
      if (i > 0)
	[s print ", "];
      [self[i] write s];
    }

  = [s print ')'];
}

end;

/******************** Type ********************/

implementation class
Type extension tm
end;

implementation instance
Type extension tm

OutputStream
  outputType Output o
	  to OutputStream s
{
  = [o outputType self];
}

end;

/******************** TypeTuple ********************/

<doc> Extension to {TypeTuple} for interfacing with {Output}.  </doc>
implementation class
TypeTuple extension tm
end;

implementation instance
TypeTuple extension tm

redefine OutputStream
  outputType Output o
	  to OutputStream s
{
  int i;

  [s print '('];
  for (i = 0; i < length; i++)
    {
      if (i > 0)
	[s print ", "];
      [self[i] outputType o to s];
    }

  = [s print ')'];
}

end;

/******************** UnresolvedInvocation ********************/

implementation class
UnresolvedInvocation extension tm

end;

implementation instance
UnresolvedInvocation extension tm

OutputStream
  write OutputStream s
{
  int i, nn = [name_parts length], na = !arguments ? 0 : [arguments length];

  [s print ('[', receiver)];

  for (i = 0; i < nn; i++)
    {
      String np = name_parts[i];

      [s print (' ', np)];
      if (i < na)
	{
	  Expression e = arguments[i];
	  [s print (' ', e)];
	}
    }

  = [s print ']'];
}

end;

/******************** UnresolvedOperatorInvocation ********************/

implementation class
UnresolvedOperatorInvocation extension tm

OutputStream
  write_argument (OutputStream, Argument, int) (s, arg, priority)
{
  if (!arg)
    s = [s print arg];
  else
    s = [arg write s priority priority];
  = s;
}

end;

implementation instance
UnresolvedOperatorInvocation extension tm

OutputStream
  write OutputStream s
{
  = [self write s priority -1];
}

OutputStream
     write OutputStream s
  priority int priority
{
  int n = [arguments length];

  if (priority > operator)
    s = [s print '('];

  if (n == 1)
    {
      s = [operator_names[operator] write s];
      if (operator == OP_OLD || operator == OP_RET)
	s = [s print ' '];
      write_argument (s, arguments[0], operator);
    }
  else if (n == 2)
    {
      write_argument (s, arguments[0], operator);
      if (operator == OP_AT)
	{
	  s = [s print '['];
	  write_argument (s, arguments[1], operator);
	  s = [s print ']'];
	}
      else
	{
	  s = [s print ' '];
	  s = [operator_names[operator] write s];
	  s = [s print ' '];
	  write_argument (s, arguments[1], operator);
	}
    }
  else
    {
      write_argument (s, arguments[0], operator);
      s = [" ? " write s];
      write_argument (s, arguments[1], operator);
      s = [" : " write s];
      write_argument (s, arguments[2], operator);
    }

  if (priority > operator)
    s = [s print ')'];

  = s;
}

end;

/******************** VariableSingle ********************/

implementation class
VariableSingle extension tm

end;

implementation instance
VariableSingle extension tm

OutputStream
  outputType Output o
	  to OutputStream s
{
  [o outputType type];
}

end;
